from .. import units as pq
from .common import TestCase
import numpy as np

class TestQuantityMethods(TestCase):

    def setUp(self):
        self.q = [[1, 2], [3, 4]] * pq.m

    def test_tolist(self):
        self.assertEqual(self.q.tolist(), [[1*pq.m, 2*pq.m], [3*pq.m, 4*pq.m]])
        q_singleton = 1 * pq.m
        self.assertEqual(q_singleton.tolist(), q_singleton)

    def test_sum(self):
        self.assertQuantityEqual(self.q.sum(), 10*pq.m)
        self.assertQuantityEqual(self.q.sum(0), [4, 6]*pq.m)
        self.assertQuantityEqual(self.q.sum(1), [3, 7]*pq.m)

    def test_nansum(self):
        import numpy as np
        qnan = [[1,2], [3,4], [np.nan,np.nan]] * pq.m
        self.assertQuantityEqual(qnan.nansum(), 10*pq.m )
        self.assertQuantityEqual(qnan.nansum(0), [4,6]*pq.m )

    def test_fill(self):
        self.q.fill(6 * pq.ft)
        self.assertQuantityEqual(self.q, [[6, 6], [6, 6]] * pq.ft)
        self.q.fill(5)
        self.assertQuantityEqual(self.q, [[5, 5], [5, 5]] * pq.ft)

    def test_reshape(self):
        self.assertQuantityEqual(self.q.reshape([1,4]), [[1, 2, 3, 4]] * pq.m)

    def test_transpose(self):
        self.assertQuantityEqual(self.q.transpose(), [[1, 3], [2, 4]] * pq.m)

    def test_flatten(self):
        self.assertQuantityEqual(self.q.flatten(), [1, 2, 3, 4] * pq.m)

    def test_ravel(self):
        self.assertQuantityEqual(self.q.ravel(), [1, 2, 3, 4] * pq.m)

    def test_squeeze(self):
        self.assertQuantityEqual(
            self.q.reshape([1,4]).squeeze(),
            [1, 2, 3, 4] * pq.m
            )

    def test_take(self):
        self.assertQuantityEqual(self.q.take([0,1,2,3]), self.q.flatten())

    def test_put(self):
        q = self.q.flatten()
        q.put([0,2], [10,20]*pq.m)
        self.assertQuantityEqual(q, [10, 2, 20, 4]*pq.m)

        q = self.q.flatten()
        q.put([0, 2], [1, 2]*pq.mm)
        self.assertQuantityEqual(q, [0.001, 2, 0.002, 4]*pq.m)

        q = self.q.flatten()/pq.mm
        q.put([0, 2], [1, 2])
        self.assertQuantityEqual(q.simplified, [1, 2000, 2, 4000])
        self.assertQuantityEqual(q, [0.001, 2, 0.002, 4]*pq.m/pq.mm)

        q = self.q.flatten()
        self.assertRaises(ValueError, q.put, [0, 2], [4, 6] * pq.J)
        self.assertRaises(ValueError, q.put, [0, 2], [4, 6])

    def test_repeat(self):
        self.assertQuantityEqual(self.q.repeat(2), [1,1,2,2,3,3,4,4]*pq.m)

    def test_sort(self):
        q = [4, 5, 2, 3, 1, 6] * pq.m
        q.sort()
        self.assertQuantityEqual(q, [1, 2, 3, 4, 5, 6] * pq.m)

    def test_argsort(self):
        q = [1, 4, 5, 6, 2, 9] * pq.MeV
        self.assertQuantityEqual(q.argsort(), [0, 4, 1, 2, 3, 5])

    def test_diagonal(self):
        q = [[1, 2, 3], [1, 2, 3], [1, 2, 3]] * pq.m
        self.assertQuantityEqual(q.diagonal(offset=1), [2, 3] * pq.m)

    def test_compress(self):
        self.assertQuantityEqual(
            self.q.compress([False, True], axis=0),
            [[3, 4]] * pq.m
            )
        self.assertQuantityEqual(
            self.q.compress([False, True], axis=1),
            [[2], [4]] * pq.m
            )

    def test_searchsorted(self):
        self.assertQuantityEqual(
            self.q.flatten().searchsorted([1.5, 2.5] * pq.m),
            [1, 2]
            )

        self.assertRaises(ValueError, self.q.flatten().searchsorted, [1.5, 2.5])

    def test_nonzero(self):
        q = [1, 0, 5, 6, 0, 9] * pq.m
        self.assertQuantityEqual(q.nonzero()[0], [0, 2, 3, 5])

    def methodWithOut(self, name, result, q=None, *args, **kw):
        import numpy as np
        from .. import Quantity

        if q is None:
            q = self.q

        self.assertQuantityEqual(
            getattr(q.copy(), name)(*args,**kw),
            result
        )
        if isinstance(result, Quantity):
            # deliberately using an incompatible unit
            out = Quantity(np.empty_like(result.magnitude), pq.s, copy=False)
        else:
            out = np.empty_like(result)
        ret = getattr(q.copy(), name)(*args, out=out, **kw)
        self.assertQuantityEqual(
            ret,
            result
        )
        # returned array should be the same as out
        self.assertEqual(id(ret),id(out))
        # but the units had to be adjusted
        if isinstance(result, Quantity):
            self.assertEqual(ret.units,result.units)
        else:
            self.assertEqual(
                getattr(ret, 'units', pq.dimensionless),
                pq.dimensionless
            )


    def test_max(self):
        self.methodWithOut('max', 4 * pq.m)
        self.methodWithOut('max', [3, 4] * pq.m, axis=0)
        self.methodWithOut('max', [2, 4] * pq.m, axis=1)

    def test_nanmax(self):
        q = np.append(self.q, np.nan) * self.q.units
        self.assertQuantityEqual(q.nanmax(), 4*pq.m)

    def test_argmax(self):
        import numpy as np
        self.assertQuantityEqual(self.q.argmax(), 3)
        self.assertQuantityEqual(self.q.argmax(axis=0), [1, 1])
        self.assertQuantityEqual(self.q.argmax(axis=1), [1, 1])
        # apparently, numpy's argmax does not return the same object when out is specified.
        # instead, we test here for shared data
        out = np.r_[0, 0]
        ret = self.q.argmax(axis=0,out=out)
        self.assertQuantityEqual(ret, [1, 1])
        self.assertEqual(ret.ctypes.data, out.ctypes.data)

    def test_nanargmax(self):
        q = np.append(self.q, np.nan) * self.q.units
        self.assertEqual(self.q.nanargmax(), 3)

    def test_min(self):
        self.methodWithOut('min', 1 * pq.m)
        self.methodWithOut('min', [1, 2] * pq.m, axis=0)
        self.methodWithOut('min', [1, 3] * pq.m, axis=1)

    def test_nanmin(self):
        q = np.append(self.q, np.nan) * self.q.units
        self.assertQuantityEqual(q.nanmin(), 1*pq.m)

    def test_argmin(self):
        import numpy as np
        self.assertQuantityEqual(self.q.argmin(), 0)
        self.assertQuantityEqual(self.q.argmin(axis=0), [0, 0])
        self.assertQuantityEqual(self.q.argmin(axis=1), [0, 0])
        # apparently, numpy's argmax does not return the same object when out is specified.
        # instead, we test here for shared data
        out = np.r_[2, 2]
        ret = self.q.argmin(axis=0,out=out)
        self.assertQuantityEqual(ret, [0, 0])
        self.assertEqual(ret.ctypes.data, out.ctypes.data)

    def test_nanargmax(self):
        q = np.append(self.q, np.nan) * self.q.units
        self.assertEqual(self.q.nanargmin(), 0)
        
    def test_ptp(self):
        self.methodWithOut('ptp', 3 * pq.m)
        self.methodWithOut('ptp', [2, 2] * pq.m, axis=0)
        self.methodWithOut('ptp', [1, 1] * pq.m, axis=1)

    def test_clip(self):
        self.methodWithOut(
            'clip',
            [[1, 2], [2, 2]] * pq.m,
            max=2*pq.m,
        )
        self.methodWithOut(
            'clip',
            [[3, 3], [3, 4]] * pq.m,
            min=3*pq.m,
        )
        self.methodWithOut(
            'clip',
            [[2, 2], [3, 3]] * pq.m,
            min=2*pq.m, max=3*pq.m
        )
        self.assertRaises(ValueError, self.q.clip, pq.J)
        self.assertRaises(ValueError, self.q.clip, 1)

    def test_round(self):
        q = [1, 1.33, 5.67, 22] * pq.m
        self.methodWithOut(
            'round',
            [1, 1, 6, 22] * pq.m,
            q=q,
            decimals=0,
        )
        self.methodWithOut(
            'round',
            [0, 0, 10, 20] * pq.m,
            q=q,
            decimals=-1,
        )
        self.methodWithOut(
            'round',
            [1, 1.3, 5.7, 22] * pq.m,
            q=q,
            decimals=1,
        )

    def test_trace(self):
        self.methodWithOut('trace', (1+4) * pq.m)

    def test_cumsum(self):
        self.methodWithOut('cumsum', [1, 3, 6, 10] * pq.m)
        self.methodWithOut('cumsum', [[1, 2], [4, 6]] * pq.m, axis=0)
        self.methodWithOut('cumsum', [[1, 3], [3, 7]] * pq.m, axis=1)

    def test_mean(self):
        self.methodWithOut('mean', 2.5 * pq.m)
        self.methodWithOut('mean', [2, 3] * pq.m, axis=0)
        self.methodWithOut('mean', [1.5, 3.5] * pq.m, axis=1)

    def test_nanmean(self):
        import numpy as np    
        q = [[1,2], [3,4], [np.nan,np.nan]] * pq.m
        self.assertQuantityEqual(q.nanmean(), self.q.mean())

    def test_var(self):
        self.methodWithOut('var', 1.25 * pq.m**2)
        self.methodWithOut('var', [1, 1] * pq.m**2, axis=0)
        self.methodWithOut('var', [0.25, 0.25] * pq.m**2, axis=1)

    def test_std(self):
        self.methodWithOut('std', 1.1180339887498949 * pq.m)
        self.methodWithOut('std', [1, 1] * pq.m, axis=0)
        self.methodWithOut('std', [0.5, 0.5] * pq.m, axis=1)

    def test_nanstd(self):
        import numpy as np    
        q0 = [[1,2], [3,4]] * pq.m
        q1 = [[1,2], [3,4], [np.nan,np.nan]] * pq.m
        self.assertQuantityEqual(q0.std(), q1.nanstd())

    def test_prod(self):
        self.methodWithOut('prod', 24 * pq.m**4)
        self.methodWithOut('prod', [3, 8] * pq.m**2, axis=0)
        self.methodWithOut('prod', [2, 12] * pq.m**2, axis=1)

    def test_cumprod(self):
        self.assertRaises(ValueError, self.q.cumprod)
        self.assertQuantityEqual((self.q/pq.m).cumprod(), [1, 2, 6, 24])
        q = self.q/pq.m
        self.methodWithOut(
            'cumprod',
            [1, 2, 6, 24],
            q=q,
        )
        self.methodWithOut(
            'cumprod',
            [[1, 2], [3, 8]],
            q=q,
            axis=0,
        )
        self.methodWithOut(
            'cumprod',
            [[1, 2], [3, 12]],
            q=q,
            axis=1,
        )

    def test_conj(self):
        self.assertQuantityEqual((self.q*(1+1j)).conj(), self.q*(1-1j))
        self.assertQuantityEqual((self.q*(1+1j)).conjugate(), self.q*(1-1j))

    def test_real(self):
        test_q = self.q * (1 + 1j)
        test_q.real = [[39.3701, 39.3701], [39.3701, 39.3701]] * pq.inch
        self.assertQuantityEqual(test_q.real, [[1., 1.], [1., 1.]] * pq.m)

    def test_imag(self):
        test_q = self.q * (1 + 1j)
        test_q.imag = [[39.3701, 39.3701], [39.3701, 39.3701]] * pq.inch
        self.assertQuantityEqual(test_q.imag, [[1., 1.], [1., 1.]] * pq.m)

    def test_getitem(self):
        self.assertRaises(IndexError, self.q.__getitem__, (0,10))
        self.assertQuantityEqual(self.q[0], [1,2]*pq.m)
        self.assertQuantityEqual(self.q[1,1], 4*pq.m)

    def test_setitem (self):
        self.assertRaises(ValueError, self.q.__setitem__, (0,0), 1)
        self.assertRaises(ValueError, self.q.__setitem__, (0,0), 1*pq.J)
        self.assertRaises(ValueError, self.q.__setitem__, 0, 1)
        self.assertRaises(ValueError, self.q.__setitem__, 0, [1, 2])
        self.assertRaises(ValueError, self.q.__setitem__, 0, 1*pq.J)

        q = self.q.copy()
        q[0] = 1*pq.m
        self.assertQuantityEqual(q, [[1,1],[3,4]]*pq.m)

        q[0] = (1,2)*pq.m
        self.assertQuantityEqual(q, self.q)

        q[:] = 1*pq.m
        self.assertQuantityEqual(q, [[1,1],[1,1]]*pq.m)

        # check and see that dimensionless numbers work correctly
        q = [0,1,2,3]*pq.dimensionless
        q[0] = 1
        self.assertQuantityEqual(q, [1,1,2,3])
        q[0] = pq.m/pq.mm
        self.assertQuantityEqual(q, [1000, 1,2,3])

        q = [0,1,2,3] * pq.m/pq.mm
        q[0] = 1
        self.assertQuantityEqual(q, [0.001,1,2,3]*pq.m/pq.mm)

    def test_iterator(self):
        for q in self.q.flatten():
            self.assertQuantityEqual(q.units, pq.m)



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


m_t = triton_mass = UnitConstant(
    'triton_mass',
    _cd('triton mass'),
    symbol='m_t'
)
g_t = triton_g_factor = UnitConstant(
    'triton_g_factor',
    _cd('triton g factor'),
    symbol='g_t'
)
mu_t = triton_magnetic_moment = UnitConstant(
    'triton_magnetic_moment',
    _cd('triton magnetic moment'),
    symbol='mu_t',
    u_symbol='μ_t'
)
triton_mass_energy_equivalent = UnitConstant(
    'triton_mass_energy_equivalent',
    _cd('triton mass energy equivalent'),
    symbol='(m_t*c**2)',
    u_symbol='(m_t·c²)'
)
triton_mass_energy_equivalent_in_MeV = UnitConstant(
    'triton_mass_energy_equivalent_in_MeV',
    _cd('triton mass energy equivalent in MeV')
)
triton_mass_in_u = UnitConstant(
    'triton_mass_in_u',
    _cd('triton mass in u')
)
triton_molar_mass = UnitConstant(
    'triton_molar_mass',
    _cd('triton molar mass'),
    symbol='M_t'
)

triton_electron_mass_ratio = UnitConstant(
    'triton_electron_mass_ratio',
    _cd('triton-electron mass ratio'),
    symbol='(m_t/m_e)',
    u_symbol='(m_t/mₑ)'
)
triton_proton_mass_ratio = UnitConstant(
    'triton_proton_mass_ratio',
    _cd('triton-proton mass ratio'),
    symbol='(m_t/m_p)',
    u_symbol='(m_t/m_p)'
)

triton_electron_magnetic_moment_ratio = UnitConstant(
    'triton_electron_magnetic_moment_ratio',
    _cd('triton-electron magnetic moment ratio'),
    symbol='(mu_t/mu_e)',
    u_symbol='(μ_t/μₑ)'
)
triton_magnetic_moment_to_Bohr_magneton_ratio = UnitConstant(
    'triton_magnetic_moment_to_Bohr_magneton_ratio',
    _cd('triton magnetic moment to Bohr magneton ratio'),
    symbol='(mu_t/mu_B)',
    u_symbol='(μ_t/μ_B)'
)
triton_magnetic_moment_to_nuclear_magneton_ratio = UnitConstant(
    'triton_magnetic_moment_to_nuclear_magneton_ratio',
    _cd('triton magnetic moment to nuclear magneton ratio'),
    symbol='(mu_t/mu_N)',
    u_symbol='(μ_t/μ_N)'
)
triton_neutron_magnetic_moment_ratio = UnitConstant(
    'triton_neutron_magnetic_moment_ratio',
    _cd('triton-neutron magnetic moment ratio'),
    symbol='(mu_t/mu_n)',
    u_symbol='(μ_t/μ_n)'
)
triton_proton_magnetic_moment_ratio = UnitConstant(
    'triton_proton_magnetic_moment_ratio',
    _cd('triton-proton magnetic moment ratio'),
    symbol='(mu_t/mu_p)',
    u_symbol='(μ_t/μ_p)'
)

del UnitConstant, _cd



"""
"""


from . import prefixes
from .prefixes import *

from . import acceleration
from .acceleration import *

from . import angle
from .angle import *

from . import area
from .area import *

from . import compound
from .compound import *

from . import concentration
from .concentration import *

from . import dimensionless
from .dimensionless import *

from . import electromagnetism
from .electromagnetism import *

from . import energy
from .energy import *

from . import force
from .force import *

from . import frequency
from .frequency import *

from . import heat
from .heat import *

from . import information
from .information import *

from . import length
from .length import *

from . import mass
from .mass import *

from . import power
from .power import *

from . import pressure
from .pressure import *

from . import radiation
from .radiation import *

from . import substance
from .substance import *

from . import temperature
from .temperature import *

from . import time
from .time import *

from . import velocity
from .velocity import *

from . import viscosity
from .viscosity import *

from . import volume
from .volume import *

from ..unitquantity import set_default_units



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


N_A = L = Avogadro_constant = UnitConstant(
    'Avogadro_constant',
    _cd('Avogadro constant'),
    symbol='N_A'
)
n_0 = Loschmidt_constant = UnitConstant(
    'Loschmidt_constant',
    _cd('Loschmidt constant (273.15 K, 101.325 kPa)'),
    symbol='n_0',
    u_symbol='n₀'
)
R = molar_gas_constant = UnitConstant(
    'molar_gas_constant',
    _cd('molar gas constant'),
    symbol='R'
)

k = Boltzmann_constant = UnitConstant(
    'Boltzmann_constant',
    _cd('Boltzmann constant'),
    symbol='k'
)
Boltzmann_constant_in_eV_per_K = UnitConstant(
    'Boltzmann_constant_in_eV_per_K',
    _cd('Boltzmann constant in eV/K')
)
Boltzmann_constant_in_Hz_per_K = UnitConstant(
    'Boltzmann_constant_in_Hz_per_K',
    _cd('Boltzmann constant in Hz/K')
)
Boltzmann_constant_in_inverse_meters_per_kelvin = UnitConstant(
    'Boltzmann_constant_in_inverse_meters_per_kelvin',
    _cd('Boltzmann constant in inverse meters per kelvin')
)

M_u = molar_mass_constant = UnitConstant(
    'molar_mass_constant',
    _cd('molar mass constant'),
    symbol='M_u',
    u_symbol='Mᵤ'
)
molar_volume_of_ideal_gas_ST_100kPa = UnitConstant(
    'molar_volume_of_ideal_gas_ST_100kPa',
    _cd('molar volume of ideal gas (273.15 K, 100 kPa)')
)
molar_volume_of_ideal_gas_STP = UnitConstant(
    'molar_volume_of_ideal_gas_STP',
    _cd('molar volume of ideal gas (273.15 K, 101.325 kPa)')
)
molar_volume_of_silicon = UnitConstant(
    'molar_volume_of_silicon',
    _cd('molar volume of silicon')
)

del UnitConstant, _cd



import operator as op

from .. import units as pq
from ..dimensionality import Dimensionality
from .common import TestCase

meter = Dimensionality({pq.m: 1})
meter_str = 'm'
centimeter = Dimensionality({pq.cm: 1})
centimeter_str = 'cm'
joule = Dimensionality({pq.kg: 1, pq.m: 2, pq.s: -2})
joule_str = 'kg*m**2/s**2'
joule_uni = 'kg·m²/s²'
joule_tex = r'$\mathrm{\frac{kg{\cdot}m^{2}}{s^{2}}}$'
joule_htm = 'kg&sdot;m<sup>2</sup>/s<sup>2</sup>'
Joule = Dimensionality({pq.J: 1})
Joule_str = 'J'

class TestDimensionality(TestCase):

    def test_dimensionality_str(self):
        self.assertEqual(str(meter), meter_str)
        self.assertEqual(joule.string, joule_str)
        self.assertEqual(joule.unicode, joule_uni)
        self.assertEqual(joule.latex, joule_tex)
        self.assertEqual(joule.html, joule_htm)
        self.assertEqual(Joule.string, 'J')

    def test_equality(self):
        self.assertTrue(meter == meter)
        self.assertTrue(joule == joule)
        self.assertFalse(meter == Joule)
        self.assertFalse(joule == Joule)

    def test_inequality(self):
        self.assertFalse(meter != meter)
        self.assertFalse(joule != joule)
        self.assertTrue(meter != Joule)
        self.assertTrue(joule != Joule)

    def test_copy(self):
        temp = meter.copy()
        self.assertTrue(temp is not meter)
        self.assertTrue(isinstance(temp, Dimensionality))
        self.assertTrue(temp == meter)
        temp[pq.m] += 1
        self.assertFalse(temp == meter)

    def test_addition(self):
        self.assertTrue(meter + meter is not meter)
        self.assertRaises(ValueError, op.add, meter, joule)
        self.assertRaises(ValueError, op.add, Joule, joule)
        self.assertRaises(TypeError, op.add, Joule, 0)
        self.assertRaises(TypeError, op.add, 0, joule)

    def test_inplace_addition(self):
        temp = meter.copy()
        temp += meter
        self.assertEqual(temp, meter)
        self.assertRaises(ValueError, op.iadd, meter, joule)
        self.assertRaises(ValueError, op.iadd, Joule, joule)
        self.assertRaises(TypeError, op.iadd, Joule, 0)
        self.assertRaises(TypeError, op.iadd, 0, joule)

    def test_subtraction(self):
        self.assertTrue(meter - meter is not meter)
        self.assertRaises(ValueError, op.sub, meter, joule)
        self.assertRaises(ValueError, op.sub, Joule, joule)
        self.assertRaises(TypeError, op.sub, Joule, 0)
        self.assertRaises(TypeError, op.sub, 0, joule)

    def test_inplace_subtraction(self):
        temp = meter.copy()
        temp -= meter
        self.assertEqual(temp, meter)
        self.assertRaises(ValueError, op.isub, meter, joule)
        self.assertRaises(ValueError, op.isub, Joule, joule)
        self.assertRaises(TypeError, op.isub, Joule, 0)
        self.assertRaises(TypeError, op.isub, 0, joule)

    def test_multiplication(self):
        self.assertEqual(meter*meter, Dimensionality({pq.m: 2}))
        self.assertEqual(meter*centimeter, Dimensionality({pq.m: 1, pq.cm: 1}))
        self.assertEqual(joule*meter, Dimensionality({pq.kg: 1, pq.m: 3, pq.s: -2}))
        self.assertRaises(TypeError, op.mul, Joule, 0)
        self.assertRaises(TypeError, op.mul, 0, joule)

    def test_inplace_multiplication(self):
        temp = meter.copy()
        temp *= meter
        self.assertEqual(temp, meter*meter)
        temp *= centimeter
        self.assertEqual(temp, meter*meter*centimeter)
        temp *= centimeter**-1
        self.assertEqual(temp, meter*meter)
        self.assertRaises(TypeError, op.imul, Joule, 0)
        self.assertRaises(TypeError, op.imul, 0, joule)

    def test_division(self):
        self.assertEqual(meter/meter, Dimensionality())
        self.assertEqual(joule/meter, Dimensionality({pq.kg: 1, pq.m: 1, pq.s: -2}))
        self.assertRaises(TypeError, op.truediv, Joule, 0)
        self.assertRaises(TypeError, op.truediv, 0, joule)

    def test_inplace_division(self):
        temp = meter.copy()
        temp /= meter
        self.assertEqual(temp, meter/meter)
        temp /= centimeter
        self.assertEqual(temp, meter/meter/centimeter)
        temp /= centimeter**-1
        self.assertEqual(temp, meter/meter)
        self.assertRaises(TypeError, op.itruediv, Joule, 0)
        self.assertRaises(TypeError, op.itruediv, 0, joule)

    def test_power(self):
        self.assertEqual(meter**2, meter*meter)
        self.assertEqual(meter**0, Dimensionality())
        self.assertEqual(joule**2, Dimensionality({pq.kg: 2, pq.m: 4, pq.s: -4}))
        self.assertRaises(TypeError, op.pow, Joule, joule)
        self.assertRaises(TypeError, op.pow, joule, Joule)
        self.assertEqual(meter**-1 == meter**-2, False)

    def test_inplace_power(self):
        temp = meter.copy()
        temp **= 2
        self.assertEqual(temp, meter**2)
        temp = joule.copy()
        temp **= 2
        self.assertEqual(temp, joule**2)
        temp = meter.copy()
        temp **= 0
        self.assertEqual(temp, Dimensionality())
        self.assertRaises(TypeError, op.ipow, Joule, joule)
        self.assertRaises(TypeError, op.ipow, joule, Joule)

    def test_simplification(self):
        self.assertEqual(Joule.simplified.string, 'kg*m**2/s**2')
        self.assertEqual(Joule.simplified, joule)


    def test_gt(self):
        self.assertTrue(joule > meter)
        self.assertTrue(Joule > meter)
        self.assertFalse(meter > joule)
        self.assertFalse(meter > Joule)
        self.assertFalse(joule > joule)
        self.assertFalse(joule > Joule)

    def test_ge(self):
        self.assertTrue(joule >= meter)
        self.assertTrue(Joule >= meter)
        self.assertFalse(meter >= joule)
        self.assertFalse(meter >= Joule)
        self.assertTrue(joule >= joule)
        self.assertTrue(joule >= Joule)

    def test_lt(self):
        self.assertTrue(meter < joule)
        self.assertTrue(meter < Joule)
        self.assertFalse(joule < meter)
        self.assertFalse(Joule < meter)
        self.assertFalse(joule < joule)
        self.assertFalse(Joule < joule)

    def test_le(self):
        self.assertTrue(meter <= joule)
        self.assertTrue(meter <= Joule)
        self.assertFalse(joule <= meter)
        self.assertFalse(Joule <= meter)
        self.assertTrue(joule <= joule)
        self.assertTrue(joule <= Joule)



"""
"""

from ..unitquantity import UnitQuantity, UnitInformation, dimensionless
from .time import s

bit = UnitInformation(
    'bit',
    aliases=['bits']
)
B = byte = o = octet = UnitInformation(
    'byte',
    8*bit,
    symbol='B',
    aliases=['bytes', 'o', 'octet', 'octets']
)
kB = kilobyte = ko = UnitInformation(
    'kilobyte',
    1000 * byte,
    symbol='kB',
    aliases=['kilobytes', 'kilooctet', 'kilooctets']
)
MB = megabyte = Mo = UnitInformation(
    'megabyte',
    1000 * kilobyte,
    symbol='MB',
    aliases=['megabytes', 'megaoctet', 'megaoctets']
)
GB = gigabyte = Go = UnitInformation(
    'gigabyte',
    1000 * megabyte,
    symbol='GB',
    aliases=['gigabytes', 'gigaoctet', 'gigaoctets']
)
TB = terabyte = To = UnitInformation(
    'terabyte',
    1000 * gigabyte,
    symbol='TB',
    aliases=['terabytes', 'teraoctet', 'teraoctets']
)
PB = petabyte = Po = UnitInformation(
    'petabyte',
    1000 * terabyte,
    symbol='PB',
    aliases=['petabytes', 'petaoctet', 'petaoctets']
)
EB = exabyte = Eo = UnitInformation(
    'exabyte',
    1000 * petabyte,
    symbol='EB',
    aliases=['exabytes', 'exaoctet', 'exaoctets']
)
ZB = zettabyte = Zo = UnitInformation(
    'zettabyte',
    1000 * exabyte,
    symbol='ZB',
    aliases=['zettabytes', 'zettaoctet', 'zettaoctets']
)
YB = yottabyte = Yo = UnitInformation(
    'yottabyte',
    1000 * zettabyte,
    symbol='YB',
    aliases=['yottabytes', 'yottaoctet', 'yottaoctets']
)
Bd = baud = bps = UnitQuantity(
    'baud',
    bit/s,
    symbol='Bd',
)

# IEC
KiB = kibibyte = Kio = UnitInformation(
    'kibibyte',
    1024 * byte,
    symbol='KiB',
    aliases=['kibibytes', 'kibioctet', 'kibioctets']
)
MiB = mebibyte = Mio = UnitInformation(
    'mebibyte',
    1024 * kibibyte,
    symbol='MiB',
    aliases=['mebibytes', 'mebioctet', 'mebioctets']
)
GiB = gibibyte = Gio = UnitInformation(
    'gibibyte',
    1024 * mebibyte,
    symbol='GiB',
    aliases=['gibibytes', 'gibioctet', 'gibioctets']
)
TiB = tebibyte = Tio = UnitInformation(
    'tebibyte',
    1024 * gibibyte,
    symbol='TiB',
    aliases=['tebibytes', 'tebioctet', 'tebioctets']
)
PiB = pebibyte = Pio = UnitInformation(
    'pebibyte',
    1024 * tebibyte,
    symbol='PiB',
    aliases=['pebibytes', 'pebioctet', 'pebioctets']
)
EiB = exbibyte = Eio = UnitInformation(
    'exbibyte',
    1024 * pebibyte,
    symbol='EiB',
    aliases=['exbibytes', 'exbioctet', 'exbioctets']
)
ZiB = zebibyte = Zio = UnitInformation(
    'zebibyte',
    1024 * exbibyte,
    symbol='ZiB',
    aliases=['zebibytes', 'zebioctet', 'zebioctets']
)
YiB = yobibyte = Yio = UnitInformation(
    'yobibyte',
    1024 * zebibyte,
    symbol='YiB',
    aliases=['yobibytes', 'yobioctet', 'yobioctets']
)

del UnitQuantity, s, dimensionless



from setuptools import Command, setup
from setuptools.command.build_py import build_py as _build
from setuptools.command.sdist import sdist as _sdist
from datetime import datetime


class data(Command):

    description = "Convert the NIST database of constants"
    user_options = []
    boolean_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        with open('quantities/constants/NIST_codata.txt') as f:
            data = f.read()
        data = data.split('\n')[10:-1]

        with open('quantities/constants/_codata.py', 'w') as f:
            f.write('# THIS FILE IS AUTOMATICALLY GENERATED\n')
            f.write('# ANY CHANGES MADE HERE WILL BE LOST\n')
            f.write(f'# LAST GENERATED: {datetime.now()}\n\n')
            f.write('physical_constants = {}\n\n')
            for line in data:
                name = line[:55].rstrip().replace('mag.','magnetic')
                name = name.replace('mom.', 'moment')
                val = line[55:77].replace(' ','').replace('...','')
                prec = line[77:99].replace(' ','').replace('(exact)', '0')
                unit = line[99:].rstrip().replace(' ', '*').replace('^', '**')
                d = "{'value': %s, 'precision': %s, 'units': '%s'}" \
                    %(val, prec, unit)
                f.write("physical_constants['%s'] = %s\n"%(name, d))


class sdist(_sdist):

    def run(self):
        self.run_command('data')
        _sdist.run(self)


class build(_build):

    def run(self):
        self.run_command('data')
        _build.run(self)


setup(cmdclass={"build_py": build, "sdist": sdist, "data": data})



"""
"""

from ._utils import _cd
from ..unitquantity import UnitConstant


molar_Planck_constant = UnitConstant(
    'molar_Planck_constant',
    _cd('molar Planck constant'),
    symbol='(N_A*h)',
    u_symbol='(N_A·h)'
)
molar_Planck_constant_times_c = UnitConstant(
    'molar_Planck_constant_times_c',
    _cd('molar Planck constant times c'),
    symbol='(N_A*h*c)',
    u_symbol='(N_A·h·c)'
)
h = Planck_constant = UnitConstant(
    'Planck_constant',
    _cd('Planck constant'),
    symbol='h'
)
hbar = Planck_constant_over_2_pi = UnitConstant(
    'Planck_constant_over_2_pi',
    _cd('Planck constant over 2 pi'),
    symbol='(h/(2*pi))',
    u_symbol='ħ'
)
quantum_of_circulation = UnitConstant(
    'quantum_of_circulation',
    _cd('quantum of circulation'),
    symbol='(h/(2*m_e))',
    u_symbol='(h/(2·mₑ))'
)
quantum_of_circulation_times_2 = UnitConstant(
    'quantum_of_circulation_times_2',
    _cd('quantum of circulation times 2'),
    symbol='(h/m_e)',
    u_symbol='(h/mₑ)'
)

l_P = Planck_length = UnitConstant(
    'Planck_length',
    _cd('Planck length'),
    symbol='l_P'
)
m_P = Planck_mass = UnitConstant(
    'Planck_mass',
    _cd('Planck mass'),
    symbol='m_P'
)
T_P = Planck_temperature = UnitConstant(
    'Planck_temperature',
    _cd('Planck temperature'),
    symbol='T_P'
)
t_P = Planck_time = UnitConstant(
    'Planck_time',
    _cd('Planck time'),
    symbol='t_P'
)

R_infinity = Rydberg_constant = UnitConstant(
    'Rydberg_constant',
    _cd('Rydberg constant'),
    symbol='R_infinity',
    u_symbol='R_∞'
)
Rydberg_constant_times_c_in_Hz = UnitConstant(
    'Rydberg_constant_times_c_in_Hz',
    _cd('Rydberg constant times c in Hz')
)
Rydberg_constant_times_hc_in_eV = UnitConstant(
    'Rydberg_constant_times_hc_in_eV',
    _cd('Rydberg constant times hc in eV')
)
Rydberg_constant_times_hc_in_J = UnitConstant(
    'Rydberg_constant_times_hc_in_J',
    _cd('Rydberg constant times hc in J'),
    symbol='(R_infinity*h*c)',
    u_symbol='(R_∞·h·c)'
)

G_0 = conductance_quantum = UnitConstant(
    'conductance_quantum',
    _cd('conductance quantum'),
    symbol='G_0',
    u_symbol='G₀'
)
K_J90 = conventional_value_of_Josephson_constant = UnitConstant(
    'conventional_value_of_Josephson_constant',
    _cd('conventional value of Josephson constant')
)
R_K90 = conventional_value_of_von_Klitzing_constant = UnitConstant(
    'conventional_value_of_von_Klitzing_constant',
    _cd('conventional value of von Klitzing constant')
)
Fermi_coupling_constant = UnitConstant(
    'Fermi_coupling_constant',
    _cd('Fermi coupling constant'),
    symbol='(G_F/(hbar*c)**3)',
    u_symbol='(G_F/(ħ·c)³)'
)
alpha = fine_structure_constant = UnitConstant(
    'fine_structure_constant',
    _cd('fine-structure constant'),
    symbol='alpha',
    u_symbol='α'
)
inverse_fine_structure_constant = UnitConstant(
    'inverse_fine_structure_constant',
    _cd('inverse fine-structure constant'),
    symbol='alpha**-1',
    u_symbol='α⁻¹'
)
c_1 = first_radiation_constant = UnitConstant(
    'first_radiation_constant',
    _cd('first radiation constant'),
    symbol='c_1',
    u_symbol='c₁'
)
c_1L = first_radiation_constant_for_spectral_radiance = UnitConstant(
    'first_radiation_constant_for_spectral_radiance',
    _cd('first radiation constant for spectral radiance'),
    symbol='c_1L',
    u_symbol='c₁_L'
)
inverse_of_conductance_quantum = UnitConstant(
    'inverse_of_conductance_quantum',
    _cd('inverse of conductance quantum'),
    symbol='G_0**-1',
    u_symbol='G₀⁻¹'
)
Josephson_constant = K_J = UnitConstant(
    'Josephson_constant',
    _cd('Josephson constant'),
    symbol='K_J'
)
Phi_0 = magnetic_flux_quantum = UnitConstant(
    'magnetic_flux_quantum',
    _cd('magnetic flux quantum'),
    symbol='Phi_0',
    u_symbol='Φ₀'
)
Newtonian_constant_of_gravitation_over_h_bar_c = UnitConstant(
    'Newtonian_constant_of_gravitation_over_h_bar_c',
    _cd('Newtonian constant of gravitation over h-bar c'),
    symbol='(G/(hbar*c))',
    u_symbol='(G/(ħ·c))'
)
Sackur_Tetrode_constant_ST_100kPa = UnitConstant(
    'Sackur_Tetrode_constant_ST_100kPa',
    _cd('Sackur-Tetrode constant (1 K, 100 kPa)')
)
Sackur_Tetrode_constant_STP = UnitConstant(
    'Sackur_Tetrode_constant_STP',
    _cd('Sackur-Tetrode constant (1 K, 101.325 kPa)')
)
c_2 = second_radiation_constant = UnitConstant(
    'second_radiation_constant',
    _cd('second radiation constant'),
    symbol='c_2',
    u_symbol='c₂'
)
sigma = Stefan_Boltzmann_constant = UnitConstant(
    'Stefan_Boltzmann_constant',
    _cd('Stefan-Boltzmann constant'),
    symbol='sigma',
    u_symbol='σ'
)
R_K = von_Klitzing_constant = UnitConstant(
    'von_Klitzing_constant',
    _cd('von Klitzing constant'),
    symbol='R_K'
)
b_prime = Wien_frequency_displacement_law_constant = UnitConstant(
    'Wien_frequency_displacement_law_constant',
    _cd('Wien frequency displacement law constant'),
    symbol='bprime',
    u_symbol='b′'
)
b = Wien_wavelength_displacement_law_constant = UnitConstant(
    'Wien_wavelength_displacement_law_constant',
    _cd('Wien wavelength displacement law constant'),
    symbol='b'
)

Planck_constant_in_eV_s = UnitConstant(
    'Planck_constant_in_eV_s',
    _cd('Planck constant in eV s')
)
Planck_constant_over_2_pi_in_eV_s = UnitConstant(
    'Planck_constant_over_2_pi_in_eV_s',
    _cd('Planck constant over 2 pi in eV s')
)
Planck_constant_over_2_pi_times_c_in_MeV_fm = UnitConstant(
    'Planck_constant_over_2_pi_times_c_in_MeV_fm',
    _cd('Planck constant over 2 pi times c in MeV fm')
)
Planck_mass_energy_equivalent_in_GeV = UnitConstant(
    'Planck_mass_energy_equivalent_in_GeV',
    _cd('Planck mass energy equivalent in GeV')
)

del UnitConstant, _cd



"""
"""

from ..unitquantity import dimensionless, UnitQuantity

percent = UnitQuantity(
    'percent',
    .01*dimensionless,
    symbol='%'
)

count = counts = UnitQuantity(
    'count',
    1*dimensionless,
    symbol='ct',
    aliases=['cts', 'counts']
)

lsb = UnitQuantity(
    'least_significant_bit',
    1*dimensionless,
    symbol='lsb',
    aliases=['lsbs']
)

del UnitQuantity



import sys
import unittest

import numpy as np

from ..quantity import Quantity
from ..units import set_default_units

class TestCase(unittest.TestCase):

    def setUp(self):
        set_default_units('SI')

    def tearDown(self):
        set_default_units('SI')

    def assertQuantityEqual(self, q1, q2, msg=None, delta=None):
        """
        Make sure q1 and q2 are the same quantities to within the given
        precision.
        """
        delta = 1e-5 if delta is None else delta
        msg = '' if msg is None else ' (%s)' % msg

        q1 = Quantity(q1)
        q2 = Quantity(q2)
        if q1.shape != q2.shape:
            raise self.failureException(
                f"Shape mismatch ({q1.shape} vs {q2.shape}){msg}"
                )
        if not np.all(np.abs(q1.magnitude - q2.magnitude) < delta):
            raise self.failureException(
                "Magnitudes differ by more than %g (%s vs %s)%s"
                % (delta, q1.magnitude, q2.magnitude, msg)
                )

        d1 = getattr(q1, '_dimensionality', None)
        d2 = getattr(q2, '_dimensionality', None)
        if (d1 or d2) and not (d1 == d2):
            raise self.failureException(
                f"Dimensionalities are not equal ({d1} vs {d2}){msg}"
                )



"""
"""

from ..unitquantity import UnitQuantity
from .energy import Btu, J
from .time import s, h, min
from .electromagnetism import A, V
from .length import ft
from .force import lbf


W = watt = volt_ampere = UnitQuantity(
    'watt',
    J/s,
    symbol='W',
    aliases=['watts', 'volt_ampere', 'volt_amperes', 'VA']
)
mW = milliwatt = UnitQuantity(
    'milliwatt',
    W/1000,
    symbol='mW',
    aliases=['milliwatts']
)
kW = kilowatt = UnitQuantity(
    'kilowatt',
    1000*W,
    symbol='kW',
    aliases=['kilowatts']
)
MW = megawatt = UnitQuantity(
    'megawatt',
    1000*kW,
    symbol='MW',
    aliases=['megawatts']
)
hp = horsepower = UK_horsepower = British_horsepower = UnitQuantity(
    'horsepower',
    33000*ft*lbf/min,
    symbol='hp',
    aliases=['UK_horsepower', 'British_horsepower']
)
boiler_horsepower = UnitQuantity(
    'boiler_horsepower',
    33475*Btu/h
)
metric_horsepower = UnitQuantity(
    'metric_horsepower',
    0.73549875*kW,
    doc='exact'
)
electric_horsepower = UnitQuantity(
    'electric_horsepower',
    746*W
)
water_horsepower = UnitQuantity(
    'water_horsepower',
    746.043*W,
    doc='exact'
)

refrigeration_ton = ton_of_refrigeration = UnitQuantity(
    'refrigeration_ton',
    12000*Btu/h,
    aliases=['ton_of_refrigeration']
)

del UnitQuantity, Btu, J, s, h, A, V



"""
"""

from ..unitquantity import UnitQuantity
from .length import m, nmi
from .time import s, h


c = speed_of_light = UnitQuantity(
    'speed_of_light',
    299792458*m/s,
    symbol='c',
    doc='exact'
)
kt = knot = knot_international = international_knot = UnitQuantity(
    'nautical_miles_per_hour',
    nmi/h,
    symbol='kt',
    aliases=['knot', 'knots', 'knot_international', 'international_knot']
)

del UnitQuantity, m, nmi, s, h



from .alpha import *
from .astronomy import *
from .atomicunits import *
from .deuteron import *
from .electromagnetism import *
from .electron import *
from .helion import *
from .mathematical import *
from .muon import *
from .naturalunits import *
from .neutron import *
from .proton import *
from .quantum import *
from .relationships import *
from .statisticalmechanics import *
from .tau import *
from .triton import *
from .weak import *
from .xray import *



"""
"""

from ._utils import _cd
from ..uncertainquantity import UncertainQuantity
from ..unitquantity import UnitConstant


au = astronomical_unit = UnitConstant(
    'astronomical_unit',
    UncertainQuantity(149597870700, 'm', 3),
    symbol='au',
    doc='http://en.wikipedia.org/wiki/Astronomical_unit'
)
G = Newtonian_constant_of_gravitation = UnitConstant(
    'Newtonian_constant_of_gravitation',
    _cd('Newtonian constant of gravitation'),
    symbol='G'
)

del UnitConstant, UncertainQuantity, _cd



"""
"""

import operator

import numpy as np

from . import markup
from .registry import unit_registry
from .decorators import memoize

def assert_isinstance(obj, types):
    try:
        assert isinstance(obj, types)
    except AssertionError:
        raise TypeError(
            f"arg {obj!r} must be of type {types!r}, got {type(obj)!r}"
        )


class Dimensionality(dict):

    """
    """

    @property
    def ndims(self):
        return sum(abs(i) for i in self.simplified.values())

    @property
    def simplified(self):
        if len(self):
            rq = 1*unit_registry['dimensionless']
            for u, d in self.items():
                rq = rq * u.simplified**d
            return rq.dimensionality
        else:
            return self

    @property
    def string(self):
        return markup.format_units(self)

    @property
    def unicode(self):
        return markup.format_units_unicode(self)

    @property
    def latex(self):
        return markup.format_units_latex(self)

    @property
    def html(self):
        return markup.format_units_html(self)

    def __hash__(self):
        res = hash(unit_registry['dimensionless'])
        for key in sorted(self.keys(), key=operator.attrgetter('format_order')):
            val = self[key]
            if val < 0:
                # can you believe that hash(-1)==hash(-2)?
                val -= 1
            res ^= hash((key, val))
        return res

    def __add__(self, other):
        assert_isinstance(other, Dimensionality)
        try:
            assert self == other
        except AssertionError:
            raise ValueError(
                'can not add units of %s and %s'\
                %(str(self), str(other))
            )
        return self.copy()

    __radd__ = __add__

    def __iadd__(self, other):
        assert_isinstance(other, Dimensionality)
        try:
            assert self == other
        except AssertionError:
            raise ValueError(
                'can not add units of %s and %s'\
                %(str(self), str(other))
            )
        return self

    def __sub__(self, other):
        assert_isinstance(other, Dimensionality)
        try:
            assert self == other
        except AssertionError:
            raise ValueError(
                'can not subtract units of %s and %s'\
                %(str(self), str(other))
            )
        return self.copy()

    __rsub__ = __sub__

    def __isub__(self, other):
        assert_isinstance(other, Dimensionality)
        try:
            assert self == other
        except AssertionError:
            raise ValueError(
                'can not add units of %s and %s'\
                %(str(self), str(other))
            )
        return self

    def __mul__(self, other):
        assert_isinstance(other, Dimensionality)
        new = Dimensionality(self)
        for unit, power in other.items():
            try:
                new[unit] += power
                if new[unit] == 0:
                    new.pop(unit)
            except KeyError:
                new[unit] = power
        return new

    def __imul__(self, other):
        assert_isinstance(other, Dimensionality)
        for unit, power in other.items():
            try:
                self[unit] += power
                if self[unit] == 0:
                    self.pop(unit)
            except KeyError:
                self[unit] = power
        return self

    def __truediv__(self, other):
        assert_isinstance(other, Dimensionality)
        new = Dimensionality(self)
        for unit, power in other.items():
            try:
                new[unit] -= power
                if new[unit] == 0:
                    new.pop(unit)
            except KeyError:
                new[unit] = -power
        return new

    def __itruediv__(self, other):
        assert_isinstance(other, Dimensionality)
        for unit, power in other.items():
            try:
                self[unit] -= power
                if self[unit] == 0:
                    self.pop(unit)
            except KeyError:
                self[unit] = -power
        return self

    def __pow__(self, other):
        try:
            assert np.isscalar(other)
        except AssertionError:
            raise TypeError('exponent must be a scalar, got %r' % other)
        if other == 0:
            return Dimensionality()
        new = Dimensionality(self)
        for i in new:
            new[i] *= other
        return new

    def __ipow__(self, other):
        try:
            assert np.isscalar(other)
        except AssertionError:
            raise TypeError('exponent must be a scalar, got %r' % other)
        if other == 0:
            self.clear()
            return self
        for i in self:
            self[i] *= other
        return self

    def __repr__(self):
        return 'Dimensionality({%s})' \
            % ', '.join(['%s: %s'% (u.name, e) for u, e in self.items()])

    def __str__(self):
        if markup.config.use_unicode:
            return self.unicode
        else:
            return self.string

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return hash(self) != hash(other)

    __neq__ = __ne__

    def __gt__(self, other):
        return self.ndims > other.ndims

    def __ge__(self, other):
        return self.ndims >= other.ndims

    def __lt__(self, other):
        return self.ndims < other.ndims

    def __le__(self, other):
        return self.ndims <= other.ndims

    def copy(self):
        return Dimensionality(self)


p_dict = {}

def _d_multiply(q1, q2, out=None):
    try:
        return q1._dimensionality * q2._dimensionality
    except AttributeError:
        try:
            return q1.dimensionality
        except:
            return q2.dimensionality
p_dict[np.multiply] = _d_multiply
p_dict[np.cross] = _d_multiply

def _d_divide(q1, q2, out=None):
    try:
        return q1._dimensionality / q2._dimensionality
    except AttributeError:
        try:
            return q1.dimensionality
        except:
            return q2.dimensionality**-1
p_dict[np.divide] = _d_divide
p_dict[np.true_divide] = _d_divide
p_dict[np.floor_divide] = _d_divide

def _d_check_uniform(q1, q2, out=None):
    try:
        assert q1._dimensionality == q2._dimensionality
        return q1.dimensionality
    except AssertionError:
        raise ValueError(
            'quantities must have identical units, got "%s" and "%s"' %
            (q1.units, q2.units)
        )
    except AttributeError:
        try:
            if hasattr(q1, 'dimensionality'):
                # q2 was not a quantity
                if not q1._dimensionality or not np.asarray(q2).any():
                    return q1.dimensionality
                else:
                    raise ValueError
            elif hasattr(q2, 'dimensionality'):
                # q1 was not a quantity
                if not q2._dimensionality or not np.asarray(q1).any():
                    return q2.dimensionality
                else:
                    raise ValueError
        except ValueError:
            raise ValueError(
                'quantities must have identical units, got "%s" and "%s"' %
                (q1.units, q2.units)
            )

p_dict[np.add] = _d_check_uniform
p_dict[np.subtract] = _d_check_uniform
p_dict[np.mod] = _d_check_uniform
p_dict[np.fmod] = _d_check_uniform
p_dict[np.remainder] = _d_check_uniform
p_dict[np.hypot] = _d_check_uniform
p_dict[np.equal] = _d_check_uniform
p_dict[np.not_equal] = _d_check_uniform
p_dict[np.less] = _d_check_uniform
p_dict[np.less_equal] = _d_check_uniform
p_dict[np.greater] = _d_check_uniform
p_dict[np.greater_equal] = _d_check_uniform

def _d_arctan2(q1, q2, out=None):
    try:
        assert q1._dimensionality == q2._dimensionality
        return Dimensionality()
    except AssertionError:
        raise ValueError(
            'quantities must have identical units, got "%s" and "%s"' %
            (q1.units, q2.units)
        )

p_dict[np.arctan2] = _d_arctan2

def _d_power(q1, q2, out=None):
    if getattr(q2, 'dimensionality', None):
        raise ValueError("exponent must be dimensionless")
    try:
        q2 = np.array(q2)
        p = q2.min()
        if p != q2.max():
            raise ValueError('Quantities must be raised to a uniform power')
        return q1._dimensionality**p
    except AttributeError:
        return Dimensionality()
p_dict[np.power] = _d_power

def _d_square(q1, out=None):
    return q1._dimensionality**2
p_dict[np.square] = _d_square

def _d_reciprocal(q1, out=None):
    return q1._dimensionality**-1
p_dict[np.reciprocal] = _d_reciprocal

def _d_copy(q1, out=None):
    return q1.dimensionality
p_dict[np.fabs] = _d_copy
p_dict[np.absolute] = _d_copy
p_dict[np.conjugate] = _d_copy
p_dict[np.negative] = _d_copy
p_dict[np.ones_like] = _d_copy
p_dict[np.rint] = _d_copy
p_dict[np.floor] = _d_copy
p_dict[np.fix] = _d_copy
p_dict[np.ceil] = _d_copy

def _d_clip(a1, a2, a3, q):
    return q.dimensionality
try:
    p_dict[np.core.umath.clip] = _d_clip
except AttributeError:
    pass # For compatibility with Numpy < 1.17 when clip wasn't a ufunc yet

def _d_sqrt(q1, out=None):
    return q1._dimensionality**0.5
p_dict[np.sqrt] = _d_sqrt

def _d_radians(q1, out=None):
    try:
        assert q1.units == unit_registry['degree']
    except AssertionError:
        raise ValueError(
            'expected units of degrees, got "%s"' % q1._dimensionality
        )
    return unit_registry['radian'].dimensionality
p_dict[np.radians] = _d_radians

def _d_degrees(q1, out=None):
    try:
        assert q1.units == unit_registry['radian']
    except AssertionError:
        raise ValueError(
            'expected units of radians, got "%s"' % q1._dimensionality
        )
    return unit_registry['degree'].dimensionality
p_dict[np.degrees] = _d_degrees

def _d_dimensionless(q1, out=None):
    if getattr(q1, 'dimensionality', None):
        raise ValueError("quantity must be dimensionless")
    return Dimensionality()
p_dict[np.log] = _d_dimensionless
p_dict[np.log10] = _d_dimensionless
p_dict[np.log2] = _d_dimensionless
p_dict[np.log1p] = _d_dimensionless
p_dict[np.exp] = _d_dimensionless
p_dict[np.exp2] = _d_dimensionless
p_dict[np.expm1] = _d_dimensionless
p_dict[np.logaddexp] = _d_dimensionless
p_dict[np.logaddexp2] = _d_dimensionless

def _d_trig(q1, out=None):
    try:
        assert q1.units == unit_registry['radian']
    except AssertionError:
        raise ValueError(
            'expected units of radians, got "%s"' % q1._dimensionality
        )
    return Dimensionality()
p_dict[np.sin] = _d_trig
p_dict[np.sinh] = _d_trig
p_dict[np.cos] = _d_trig
p_dict[np.cosh] = _d_trig
p_dict[np.tan] = _d_trig
p_dict[np.tanh] = _d_trig

def _d_arctrig(q1, out=None):
    if getattr(q1, 'dimensionality', None):
        raise ValueError("quantity must be dimensionless")
    return unit_registry['radian'].dimensionality
p_dict[np.arcsin] = _d_arctrig
p_dict[np.arcsinh] = _d_arctrig
p_dict[np.arccos] = _d_arctrig
p_dict[np.arccosh] = _d_arctrig
p_dict[np.arctan] = _d_arctrig
p_dict[np.arctanh] = _d_arctrig



"""
Quantities is designed to handle arithmetic and conversions of
physical quantities, which have a magnitude, dimensionality specified
by various units, and possibly an uncertainty. Quantities is designed
to work with numpy's standard ufuncs, many of which are already
supported. The package is actively developed, and while the current
features and API are stable, test coverage is incomplete and so the
package is not suggested for production use.

It is strongly suggested to import quantities to its own namespace, so
units and constants variables are not accidentally overwritten::

   >>> import quantities as pq

Here pq stands for "physical quantities" or "python quantities".
There are a number of ways to create a quantity. In practice, it is
convenient to think of quantities as a combination of a magnitude and
units. These two quantities are equivalent::

   >>> import numpy as np
   >>> q = np.array([1,2,3]) * pq.J
   >>> q = [1,2,3] * pq.J
   >>> print q
   [ 1.  2.  3.] J

The Quantity constructor can also be used to create quantities,
similar to numpy.array. Units can be designated using a string
containing standard unit abbreviations or unit names. For example::

   >>> q = pq.Quantity([1,2,3], 'J')
   >>> q = pq.Quantity([1,2,3], 'joules')

Units are also available as variables, and can be passed to
Quantity::

   >>> q = pq.Quantity([1,2,3], pq.J)

You can modify a quantity's units in place::

   >>> q = 1 * pq.m
   >>> q.units = pq.ft
   >>> print q
   3.280839895013123 ft

or equivalently::

   >>> q = 1 * pq.meter
   >>> q.units = 'ft' # or 'foot' or 'feet'
   >>> print q
   3.280839895013123 ft

Note that, with strings, units can be designated using plural
variants. Plural variants of the module variables are not available at
this time, in the interest of keeping the units namespace somewhat
manageable. `q.units = 'feet'` will work, `q.units = pq.feet` will
not.

The units themselves are special objects that can not be modified in
place::

   >>> pq.meter.units = 'feet'
   AttributeError: can not modify protected units

Instead of modifying a quantity in place, you can create a new
quantity, rescaled to the new units::

   >>> q = 300 * pq.ft * 600 * pq.ft
   >>> q2 = q.rescale('US_survey_acre')
   >>> print q2
   4.13221487605 US_survey_acre

but rescaling will fail if the requested units fails a dimensional
analysis::

   >>> q = 10 * pq.joule
   >>> q2 = q.rescale(pq.watt)
   ValueError: Unable to convert between units of "J" and "W"

Quantities can not be rescaled in place if the unit conversion fails
a dimensional analysis::

   >>> q = 10 * pq.joule
   >>> q.units = pq.watts
   ValueError: Unable to convert between units of "J" and "W"
   >>> print q
   10.0 J

Quantities will attempt to simplify units when the users intent is
unambiguous:

   >>> q = (10 * pq.meter)**3
   >>> q2 = q/(5*pq.meter)**2
   >>> print q2
   40 m

Quantities will not try to guess in an ambiguous situation:

   >>> q = (10 * pq.meter)**3
   >>> q2 = q/(5*pq.ft)**2
   >>> print q2
   40 m**3/ft**2

In that case, it is not clear whether the user wanted ft converted to
meters, or meters to feet, or neither. Instead, you can obtain a new
copy of the quantity in its irreducible units, which by default are SI
units::

   >>> q = (10 * pq.meter)**3
   >>> q2 = q/(5*pq.ft)**2
   >>> print q2
   40 m**3/ft**2
   >>> qs = q2.simplified
   >>> print qs
   430.556416668 m

It is also possible to customize the units in which simplified
quantities are expressed::

   >>> pq.set_default_units('cgs')
   >>> print pq.J.simplified
   10000000.0 g*cm**2/s**2
   >>> pq.set_default_units(length='m', mass='kg')

There are times when you may want to treat a group of units as a
single compound unit. For example, surface area per unit volume is a
fairly common quantity in materials science. If expressed in the
usual way, the quantity will be expressed in units that you may not
recognize::

   >>> q = 1 * pq.m**2 / pq.m**3
   >>> print q
   1.0 1/m

Here are some tricks for working with these compound units, which
can be preserved::

   >>> q = 1 * pq.CompoundUnit("m**2/m**3")
   >>> print q
   1.0 (m**2/m**3)

and can be simplified::

   >>> qs = q.simplified
   >>> qs
   1.0 1/m

and then rescaled back into compound units::

   >>> q2 = qs.rescale(CompoundUnit("m**2/m**3"))
   >>> print q2
   1.0 (m**2/m**3)

Compound units can be combined with regular units as well:

   >>> q = 1 * pq.CompoundUnit('parsec/cm**3') * pq.cm**2
   >>> print q
   1.0 cm**2*(parsec/cm**3)

It is easy to define a unit that is not already provided by
quantities. For example::

   >>> uK = pq.UnitQuantity('microkelvin', pq.degK/1e6, symbol='uK')
   >>> print uK
   1 uK (microkelvin)
   >>> q = 1000*uK
   >>> print q.simplified
   0.001 K

There is also support for quantities with uncertainty::

   >>> q = UncertainQuantity(4,J,.2)
   >>> q
   4.0*J
   +/-0.2*J (1 sigma)

By assuming that the uncertainties are uncorrelated, the uncertainty can be
propagated during arithmetic operations::

   >>> length = UncertainQuantity(2.0, m, .001)
   >>> width = UncertainQuantity(3.0, m, .001)
   >>> area = length*width
   >>> area
   6.0*m**2
   +/-0.00360555127546*m**2 (1 sigma)

In that case, the measurements of the length and width were independent, and
the two uncertainties presumed to be uncorrelated. Here is a warning though:

   >>> q*q
   16.0*J**2
   +/-1.1313708499*J**2 (1 sigma)

This result is probably incorrect, since it assumes the uncertainties of the two
multiplicands are uncorrelated. It would be more accurate in this case to use::

   >>> q**2
   16.0*J**2
   +/-1.6*J**2 (1 sigma)

There is an entire subpackage dedicated to physical constants. The
values of all the constants are taken from values published by the
National Institute of Standards and Technology at
http://physics.nist.gov/constants . Most physical constants have some
form of uncertainty, which has also been published by NIST. All
uncertainties are one standard deviation. There are lots of constants
and quantities includes them all (with one exception: F*, the Faraday
constant for conventional electrical current, which is defined in
units of C_90, for which I have not found a hard reference value).
Physical constants are sort of similar to compound units, for example:

   >>> print pq.constants.proton_mass
   1 m_p (proton_mass)
   >>> print pq.constants.proton_mass.simplified
   1.672621637e-27 kg
   +/-8.3e-35 kg (1 sigma)

A Latex representation of the dimensionality may be obtained in the following fashion::

    >>> g = pq.Quantity(9.80665,'m/s**2')
    >>> mass = 50 * pq.kg
    >>> weight = mass*g
    >>> print weight.dimensionality.latex
    $\\mathrm{\\frac{kg{\\cdot}m}{s^{2}}}$
    >>> weight.units = pq.N
    >>> print weight.dimensionality.latex
    $\\mathrm{N}$

The Latex output is compliant with the MathText subset used by Matplotlib.  To add
formatted units to the axis label of a Matplotlib figure, one could use::

    >>> ax.set_ylabel('Weight ' + weight.dimensionality.latex)

Greater customization is available via the markup.format_units_latex function.  It allows
the user to modify the font, the multiplication symbol, or to encapsulate the latex
string in parentheses.  Due to the complexity of CompoundUnits, the latex rendering
of CompoundUnits will utilize the latex \\frac{num}{den} construct.

Although it is not illustrated in this guide, unicode symbols can be
used to provide a more compact representation of the units. This
feature is disabled by default. It can be enabled by setting the
following in your ~/.pythonrc.py::

   quantities_unicode = True

or you can change this setting on the fly by doing::

   from quantities import markup
   markup.config.use_unicode = True # or False

Even when unicode is enabled, when you pass strings to designate
units, they should still conform to valid python expressions.

.. attention::
   Quantities is not a package for describing coordinate systems that require a
   point of reference, like positions on a map. In particular, Quantities does
   not support absolute temperature scales. Instead, temperatures are assumed to
   be temperature *differences*. For example:

   >>> T = 20 * pq.degC
   >>> print T.rescale('K')
   20.0 K

   Proper support of coordinate systems would be a fairly large undertaking and
   is outside the scope of this project.

"""


from ._version import __version__

from .registry import unit_registry

from . import quantity
from .quantity import Quantity

from . import uncertainquantity
from .uncertainquantity import UncertainQuantity

from . import unitquantity
from .unitquantity import *

from .units import *

from . import constants

from .umath import *

def test(verbosity=1):
    import unittest
    suite = unittest.TestLoader().discover('quantities')
    unittest.TextTestRunner(verbosity=verbosity).run(suite)
